# Kubetools Client
# File: kubetools/cli/wait.py
# Desc: helpers for the Kubetools command line interface

import sys

import click

from kubetools_client.cli import cli_bootstrap
from kubetools_client.cli.wait_util import wait_with_spinner


@cli_bootstrap.command()
@click.argument('build_hash')
@click.pass_obj
def wait(client, build_hash):
    '''
    Wait for a Kubetools build.
    '''

    click.echo('### Watching build: {0}'.format(
        click.style(build_hash, bold=True),
    ))
    wait_build(client, build_hash)


def wait_job(client, namespace, job_id):
    def check_job_status(previous_status):
        namespace_jobs = client.list_jobs_by_namespace(namespace=namespace)
        job = namespace_jobs[namespace][job_id]

        if (
            # Job completed the number of times we want?
            job['succeeded'] == job['completions']
            # Job has failed over X times?
            or job['failed'] >= job['backoffLimit']
        ):
            return

        return job['annotations'].get('description', 'Running job')

    wait_with_spinner(check_job_status)

    # Get the job again
    namespace_jobs = client.list_jobs_by_namespace(namespace=namespace)
    job = namespace_jobs[namespace][job_id]
    status = job['succeeded'] == job['completions']

    if status:
        click.echo('<-- Job complete, final status = {0}'.format(
            click.style('SUCCESS', 'green'),
        ))

    else:
        click.echo('<-- Job complete, final status = {0}'.format(
            click.style('ERROR', 'red'),
        ))

    if not status:
        sys.exit(1)


def wait_build(client, build_hash):
    '''
    CLI helper that waits for a build to complete while displaying a little spinner.
    '''

    def check_build_status(previous_status):
        build_data = client.get_build(build_hash)

        # If we're complete, just return nothing
        if build_data['status'] not in ('PENDING', 'RUNNING'):
            return

        status_text = build_data['status']
        if build_data['status_detail']:
            status_text = '{0}/{1}'.format(status_text, build_data['status_detail'])

        return status_text

    wait_with_spinner(check_build_status)

    # Get the build date
    build_data = client.get_build(build_hash)
    status = build_data['status']

    # We did it!
    if status == 'SUCCESS':
        click.echo('--> Namespace: {0}'.format(click.style(
            build_data['namespace'], bold=True,
        )))

        domains = build_data.get('domains')
        if domains:
            for app, domain in domains.items():
                click.echo('--> App: {0} should now be available @ {1}'.format(
                    app, click.style(domain, bold=True),
                ))

    elif status == 'ABORTED':
        click.echo('--> {0}'.format(
            click.style(
                'Aborted at stage: {0}'.format(build_data['status_detail']),
                'yellow',
            ),
        ))

    # Status should be ERROR, but any non-success here is a fail
    else:
        click.echo('--> {0}'.format(
            click.style(
                'Error at stage: {0}'.format(build_data['status_detail']),
                'red',
            ),
        ))
        click.echo()
        click.echo('--------ERROR START--------')
        click.echo(click.style(build_data.get('error', 'Unknown error'), 'yellow'))
        click.echo('---------ERROR END---------')

    if status == 'SUCCESS':
        formatted_status = click.style('SUCCESS', 'green')
    elif status == 'ABORTED':
        formatted_status = click.style('ABORTED', 'yellow')
    else:
        formatted_status = click.style('ERROR', 'red')

    click.echo('<-- Build complete, final status = {0}'.format(formatted_status))

    if status == 'ABORTED':
        sys.exit(1)

    elif status != 'SUCCESS':
        sys.exit(2)



# Kubetools Client
# File: kubetools/cli/dev_util.py
# Desc: utilities for the Kubetools development commands (dev.py)

from __future__ import print_function

import re

from subprocess import CalledProcessError, PIPE, Popen, STDOUT
from threading import Thread

import docker
import requests
import six

from pydash import memoize

from kubetools_client.cli.wait_util import UPDATE_DIVISOR, wait_with_spinner
from kubetools_client.exceptions import KubeDevCommandError, KubeDevError
from kubetools_client.log import logger
from kubetools_client.settings import get_settings

from .config import (
    create_compose_config,
    dockerise_label,
    get_all_containers,
    get_compose_filename,
    get_compose_name,
)


@memoize
def get_docker_client():
    client = docker.from_env()

    try:
        client.ping()

    except (docker.errors.APIError, requests.exceptions.ConnectionError) as e:
        raise KubeDevError((
            'Could not connect to Docker, is it running?\n'
            'Error: {0}'
        ).format(e))

    return client


def ensure_docker_dev_network():
    '''
    Ensure we have a dev Docker network for all our containers to belong to.
    '''

    docker_client = get_docker_client()

    network_names = [
        network.name
        for network in docker_client.networks.list()
    ]

    if 'dev' not in network_names:
        docker_client.networks.create('dev')


def get_all_docker_dev_network_containers():
    '''
    Gets the container names for everything using the global dev network.

    Note only containers created with the default "dev" env become part of this
    network - any custom env (eg when testing or `--env`) won't be included.
    '''

    docker_client = get_docker_client()

    labels = [
        # Ignore containers created by run/exec/etc (not services)
        'com.docker.compose.oneoff=False',
        # Load anything from a compose project
        'com.docker.compose.project',
    ]

    docker_containers = docker_client.containers.list(all=True, filters={
        'label': labels,
    })

    return docker_containers


def get_containers_status(
    kubetools_config,
    container_name=None,
    all_environments=False,
):
    '''
    Get the status of any containers for the current Kubetools project.
    '''

    docker_client = get_docker_client()

    labels = [
        # Ignore containers created by run/exec/etc (not services)
        'com.docker.compose.oneoff=False',
    ]

    if all_environments:
        # Filter by *any* docker compose - we filter out containers that don't
        # belong to this project below - this maintains compatability with
        # kubetools <8 that didn't write the kubetools.project.env label.
        labels.append('com.docker.compose.project')
    else:
        # Filter by the project/env name to only fetch this envs containers
        labels.append('com.docker.compose.project={0}'.format(
            get_compose_name(kubetools_config),
        ))

    if container_name:
        labels.append(
            'com.docker.compose.service={0}'.format(container_name),
        )

    logger.debug('Listing Docker containers with labels={0}'.format(labels))
    docker_containers = docker_client.containers.list(all=True, filters={
        'label': labels,
    })

    env_to_containers = {}
    docker_name = dockerise_label(kubetools_config['name'])

    for container in docker_containers:
        compose_project = container.labels['com.docker.compose.project']
        if all_environments:
            if not compose_project.startswith(docker_name):
                continue

            # For old Kubetools versions (<8) this label won't exist
            kubetools_name = container.labels.get('kubetools.project.name')
            if kubetools_name and kubetools_config['name'] != kubetools_name:
                continue

        env = container.labels.get('kubetools.project.env')
        # Compatability for existing containers created with kubetools <8
        if not env:
            env = compose_project.replace(docker_name, '')

        # Where the name is compose-name_container_N, get container
        name = container.name.split('_')[1]

        status = container.status == 'running'
        ports = []

        if container.attrs['NetworkSettings']['Ports']:
            for local_port, host_port in six.iteritems(
                container.attrs['NetworkSettings']['Ports'],
            ):
                if not host_port:
                    continue

                ports.append({
                    'local': local_port,
                    'host': host_port[0]['HostPort'],
                })

        container_data = {
            'up': status,
            'ports': ports,
            'id': container.id,
            'labels': container.labels,
        }

        env_containers = env_to_containers.setdefault(env, {})

        if name in env_containers:
            raise ValueError((
                'Duplicate container for env {0}!: {1}({2})'
            ).format(env, name, container_data))

        env_containers[name] = container_data

    # Always provide the current environment, even if it's empty
    current_env = kubetools_config['env']
    if current_env not in env_to_containers:
        env_to_containers[kubetools_config['env']] = {}

    for env, containers in env_to_containers.items():
        for name, container in get_all_containers(kubetools_config):
            if name not in containers:
                containers[name] = {
                    'up': None,
                    'id': None,
                    'ports': [],
                }

            containers[name]['is_dependency'] = container.get('is_dependency', False)
            containers[name]['is_deployment'] = container.get('is_deployment', False)

    if all_environments:
        return env_to_containers
    return env_to_containers.get(kubetools_config['env'], {})


def get_container_status(kubetools_config, name):
    containers = get_containers_status(kubetools_config, container_name=name)
    return containers.get(name)


def _read_command_output(command, output_lines):
    # Read the commands output indefinitely
    while True:
        stdout_line = command.stdout.readline()

        if stdout_line:
            # Remove any trailing newline
            stdout_line = stdout_line.decode().strip('\n')
            # # Strip non-alphanumeric  characters
            # stdout_line = re.sub(r'[^a-zA-Z0-9_\.\-\s]', '', stdout_line)

            # Strip any ANSI escape characters
            stdout_line = re.sub(
                r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]',
                '',
                stdout_line,
            )

            output_lines.append(stdout_line)

        # No line from the command? We're done!
        else:
            break


def _run_process_with_spinner(args):
    command = Popen(
        ' '.join(args),
        stdout=PIPE,
        stderr=STDOUT,
        bufsize=1,
        close_fds=True,
        shell=True,
    )

    # Buffers
    output_lines = []

    command_reader = Thread(
        target=_read_command_output,
        args=(command, output_lines),
    )
    command_reader.start()

    def check_status(previous_status):
        # Command complete (we've read everything)? Exit here
        if not command_reader.is_alive():
            return

        if output_lines:
            return output_lines[-1]

        return previous_status

    wait_with_spinner(
        check_status,
        # This means we run the get_line check every .5 seconds
        check_status_divisor=(UPDATE_DIVISOR / 2),
    )

    # Re-join the stdout/stderr lines
    stdout = '\n'.join(output_lines)

    # Poll the command to populate it's return code
    command.poll()

    # Ensure the command is dead
    try:
        command.terminate()
        command.kill()

    # If already dead, just ignore
    except Exception:
        pass

    return command.returncode, stdout


def run_process(args, env=None, capture_output=None):
    settings = get_settings()

    capturing_output = False

    if (
        # If we explicitly need to capture, always capture
        capture_output
        # Otherwise, capture if we're not --debug and not explicitly no capture
        or (settings.debug == 0 and capture_output is not False)
    ):
        capturing_output = True

    logger.debug('--> Executing: {0}'.format(' '.join(args)))

    try:
        # If we're capturing output - things are more complicated. We need to spawn
        # the subprocess in a thread and read its output into two lists, which we
        # then rejoin to return.
        if capturing_output:
            code, stdout = _run_process_with_spinner(args)

        # Inline? Simply start the process and "communicate", this will print stdout
        # and stderr to the terminal and also capture them into variables.
        else:
            command = Popen(args, env=env, stderr=STDOUT, close_fds=True)
            stdout, _ = command.communicate()
            code = command.returncode

        if code > 0:
            raise KubeDevCommandError(
                'Compose command failed: {0}'.format(args),
                stdout,
            )

        return stdout

    except (CalledProcessError, OSError) as e:
        raise KubeDevCommandError(
            'Compose command failed: {0}'.format(args),
            getattr(e, 'output', e),
        )


def run_compose_process(kubetools_config, command_args, **kwargs):
    # Ensure we have a compose file for this config
    create_compose_config(kubetools_config)

    compose_command = [
        'docker-compose',
        # Force us to look at the current directory, not relative to the compose
        # filename (ie .kubetools/compose-name.yml).
        '--project-directory', '.',
        # Name of the project (for the com.docker.compose.projectname label)
        '--project-name', get_compose_name(kubetools_config),
        # Filename of the YAML file to load
        '--file', get_compose_filename(kubetools_config),
    ]
    compose_command.extend(command_args)

    return run_process(compose_command, **kwargs)



from kubetools_client.log import logger


def convert_legacy_kubetools_config(config, dev):
    config.setdefault('tests', [])
    config.setdefault('upgrades', [])

    config['deployments'] = {
        config['name']: {
            'containers': {},
        },
    }

    if 'buildContexts' in config:
        config['containerContexts'] = {}

        for key, data in config.pop('buildContexts').items():
            if 'preBuildCommands' in config:
                data['preBuildCommands'] = config.get('preBuildCommands')

            config['containerContexts'][key] = {
                'build': data,
            }

    # Sort out any top level upgrades
    for upgrade in config['upgrades']:
        if 'buildContext' in upgrade:
            upgrade['containerContext'] = upgrade.pop('buildContext')

    # Sort out top level containers into a single top level deployment
    #

    for key, data in config.pop('containers').items():
        if 'buildContext' in data:
            data['containerContext'] = data.pop('buildContext')

        if dev:
            # Apply any devOverrides
            overrides = data.pop('devOverrides', None)
            if overrides:
                data.update(overrides)

            # In the old config dev-only upgrades were defined in postUpCommands for
            # each container, but now we only want them to be upgrades.
            post_up_commands = data.pop('postUpCommands', None)

            if post_up_commands:
                # Filter out postUpCommands that already exist
                existing_upgrade_commands = set()
                for upgrade in config.get('upgrades', []):
                    existing_upgrade_commands.add(tuple(upgrade['command']))

                for command in post_up_commands:
                    if tuple(command) not in existing_upgrade_commands:
                        config['upgrades'].append({
                            'command': command,
                            'containerContext': data['containerContext'],
                        })

        if 'testCommands' in data:
            for command in data.pop('testCommands'):
                config['tests'].append({
                    'command': command,
                    'containerContext': data['containerContext'],
                })

        config['deployments'][config['name']]['containers'][key] = data

    # Sort out dependencies and dev dependencies
    #

    dependencies = config.pop('dependencies', {})

    if dev:
        dev_dependencies = config.pop('devDependencies', {})

        if dev_dependencies:
            # Remove any *matching* containers from the main app - because in old
            # Kubetools devDependencies with the same name would override.
            for key in dev_dependencies.keys():
                config['deployments'][config['name']]['containers'].pop(key, None)

            dependencies.update(dev_dependencies)

    config['dependencies'] = {}

    for key, data in dependencies.items():
        if 'buildContext' in data:
            data['containerContext'] = data.pop('buildContext')

        if 'servicePorts' in data:
            data['ports'] = data['servicePorts']

        config['dependencies'][key] = {
            'containers': {
                key: data,
            },
        }

    # Sort out singletons to be deployments w/maxReplicas=1
    #

    singletons = config.pop('singletons', {})
    for key, data in singletons.items():
        if 'buildContext' in data:
            data['containerContext'] = data.pop('buildContext')

        singleton = {
            'maxReplicas': 1,
            'containers': {
                key: data,
            },
        }

        namespaces = data.get('namespaces')
        if namespaces:
            singleton.setdefault('conditions', {})['namespaces'] = namespaces

        singleton_key = '{0}-{1}'.format(config['name'], key)
        config['deployments'][singleton_key] = singleton

    logger.warning('Converted legacy Kubetools config')



# Kubetools Client
# File: kubetools/cli/lists.py
# Desc:

import click


from kubetools_client.cli import cli_bootstrap


@cli_bootstrap.group()
def job():
    '''
    Manage Kubernetes jobs.
    '''


@job.command()
@click.argument('namespace')
@click.argument('job_id')
@click.pass_obj
def abort(client, namespace, job_id):
    if client.abort_job(namespace, job_id):
        click.echo('--> Job aborted: {0}/{1}'.format(
            namespace,
            click.style(job_id, bold=True),
        ))
    else:
        click.echo(click.style('Failed to abort job!', 'red'))


@job.command()
@click.argument('namespace')
@click.argument('job_id')
@click.pass_obj
def delete(client, namespace, job_id):
    if client.delete_job(namespace, job_id):
        click.echo('--> Job deleted: {0}/{1}'.format(
            namespace,
            click.style(job_id, bold=True),
        ))
    else:
        click.echo(click.style('Failed to delete job!', 'red'))



#!/usr/bin/env python

from kubetools_client.cli import cli_bootstrap
from kubetools_client.main import run_cli

# Import click command groups
from kubetools_client.cli import generate, jobs, lists, wait  # noqa: F401, I100


run_cli(cli_bootstrap)



import json

from time import sleep

import requests
from requests.exceptions import ConnectionError

from kubetools_client.exceptions import KubeAuthError, KubeClientError
from kubetools_client.log import logger
from kubetools_client.settings import get_settings


GET_METHODS = {
    'list_manifests': 'manifests',
    'list_apps_by_namespace': ('apps', 'namespace_apps'),
    'list_jobs_by_namespace': ('jobs', 'namespace_jobs'),
    'list_builds': 'builds',
    'list_environments': 'environments',
    'list_locks': 'locks',
}


class KubeClient(object):
    def __init__(
        self,
        host=None,
        port=None,
        api_version=None,
        kube_env=None,
        api_key=None,
    ):
        settings = get_settings()

        host = host or settings.KUBETOOLS_HOST
        port = port or settings.KUBETOOLS_PORT
        api_version = api_version or settings.KUBETOOLS_API_VERSION

        kube_env = kube_env or settings.DEFAULT_KUBE_ENV

        self.base_url = 'http://{0}:{1}/api/v{2}'.format(host, port, api_version)
        self.kube_env = kube_env
        self.host = host
        self.port = port

        # Setup a requests session with any authentication methods we have configured
        session = requests.session()

        if api_key:
            session.headers['Authorization'] = api_key
        elif settings.KUBETOOLS_SESSION:
            session.cookies['session'] = settings.KUBETOOLS_SESSION

        self.session = session

    def __getattr__(self, key):
        if key in GET_METHODS:
            url = data_key = GET_METHODS[key]

            if isinstance(url, tuple):
                url, data_key = url

            def wrapper(**kwargs):
                response = self._request('GET', url, params=kwargs)
                return response.get(data_key)

            return wrapper

    def _request(self, method, endpoint, **kwargs):
        request_map = {
            'GET': self.session.get,
            'POST': self.session.post,
            'DELETE': self.session.delete,
        }

        if method not in request_map:
            raise KubeClientError('Invalid request method: {0}'.format(method))

        method_func = request_map[method]
        endpoint = '{0}/{1}'.format(self.base_url, endpoint)

        # Pop and apply any env param
        params = kwargs.pop('params', {})
        if self.kube_env:
            params['kube-env'] = self.kube_env

        logger.debug('Making requests {0} {1} ({2})'.format(
            method, endpoint, kwargs,
        ))

        try:
            response = method_func(endpoint, params=params, **kwargs)

        except ConnectionError:
            raise KubeClientError(
                'Could not connect to kubetools API: {0}'.format(endpoint),
            )

        if response.status_code == 200:
            return response.json()
        else:
            # Get the JSON error if available
            try:
                data = response.json()
                content = '{0} {1}: {2}'.format(
                    data['status_code'],
                    data['error_name'],
                    data['error_message'],
                )

            # It really broke, no JSON returned!
            except Exception:
                content = 'Invalid API response: {0}: {1}'.format(
                    response, response.content,
                )

            if response.status_code == 403:
                raise KubeAuthError(content)
            raise KubeClientError(content)

    def _json_request(self, *args, **kwargs):
        kwargs['headers'] = {
            'Content-Type': 'application/json',
        }

        if 'data' in kwargs:
            kwargs['data'] = json.dumps(kwargs['data'])

        return self._request(*args, **kwargs)

    def get_build_view_url(self, build_hash):
        '''
        Returns the frontend build URL for a given build hash.
        '''

        return 'http://{0}:{1}/build/{2}'.format(self.host, self.port, build_hash)

    def get_build(self, build_hash):
        '''
        Gets the detail of a single build.

        Args:
            build_hash (str): the hash of the build to lookup

        Returns:
            build data
        '''

        response = self._request('GET', 'build/{0}'.format(build_hash))
        return response['build']

    def wait_for_build(
        self, build_hash,
        statuses=('SUCCESS', 'ERROR'),
        sleep_time=1,
        max_errors=300,
    ):
        '''
        Wait for a build to be complete.

        Args:
            build_hash (str): the hash of the build to wait for
            statuses (tuple/list): the acceptable statuses to stop waiting on
            sleep_time (int): time to sleep between checks for complection
            max_errors (int): number of API errors in a row before giving up \
            (and raising an exception)

        Returns:
            build data
        '''

        build = self.get_build(build_hash)
        n_errors = 0

        while build['status'] not in statuses:
            sleep(sleep_time)

            try:
                build = self.get_build(build_hash)
                n_errors = 0  # reset the error count

            except KubeClientError as e:
                logger.warning('API error waiting for build: {0}'.format(e))

                if n_errors >= 300:
                    raise

                n_errors += 1

        return build

    def create_build(self, build_spec):
        '''
        Creates/schedules a build in Kubetools.

        Args:
            build_spec (dict): list of dicts containing at least field "app" and \
            optionally "replicas" and "version" where version is a Git branch

        Returns:
            the build hash
        '''

        response = self._json_request(
            'POST', 'build',
            data=build_spec,
        )

        return response['hash']

    def abort_build(self, build_hash):
        '''
        Aborts a build in Kubetools.

        Args:
            build_hash (str): the hash of the build to abort

        Returns:
            the api version
        '''

        endpoint = 'build/{0}/abort'.format(build_hash)

        response = self._request(
            'POST', endpoint,
        )

        return response

    def create_deploy(self, deploy_spec, namespace=None):
        '''
        Creates/schedules a build & deploy in Kubetools.

        Args:
            deploy_spec (list): list of dicts containing the app/replicas/version
            namespace (str): optional namespace to deploy to
        '''

        response = self._json_request(
            'POST', 'deploy',
            data=deploy_spec,
            params={'namespace': namespace},
        )

        return response['hash']

    def create_run(self, run_spec, namespace=None):
        '''
        Creates/schedules a build & run in Kubetools.

        Args:
            run_spec (list): list of dicts containing the app/replicas/version
            namespace (str): optional namespace to run to
        '''

        response = self._json_request(
            'POST', 'run',
            data=run_spec,
            params={'namespace': namespace},
        )

        return response['hash']

    def create_upgrade(self, namespace, upgrade_spec=None):
        '''
        Creates/schedules an upgrade in Kubetools.

        Args:
            namespace (str): the namespace to perform the upgrade in
            upgrade_spec (list): as above, contains a list of app/version/replicas

        Returns:
            the build hash
        '''

        response = self._json_request(
            'POST', 'upgrade/{0}'.format(namespace),
            data=upgrade_spec,
        )

        return response['hash']

    def create_remove(self, namespace, remove_spec=None):
        '''
        Creates/schedules a remove in Kubetools.

        Args:
            namespace (str): the namespace to remove apps from
            remove_spec (list): as above, contains a list of apps

        Returns:
            the build hash
        '''

        response = self._json_request(
            'POST', 'remove/{0}'.format(namespace),
            data=remove_spec,
        )

        return response['hash']

    def create_restart(self, namespace, restart_spec=None):
        '''
        Creates/schedules a restart in Kubetools.

        Args:
            namespace (str): the namespace to restart apps from
            restart_spec (list): as above, contains a list of apps

        Returns:
            the build hash
        '''

        response = self._json_request(
            'POST', 'restart/{0}'.format(namespace),
            data=restart_spec,
        )

        return response['hash']

    def create_cleanup(self, namespace):
        '''
        Creates/schedules a cleanup in Kubetools.

        Args:
            namespace (str): the namespace to cleanup

        Returns:
            the build hash
        '''

        response = self._json_request(
            'POST',
            'cleanup/{0}'.format(namespace),
        )

        return response['hash']

    def abort_job(self, namespace, job_id):
        '''
        Aborts a running Kubernetes job.

        Args:
            namespace (str): the namespace the job is running in
            job_id (str): the ID of the job to abort

        Returns:
            boolean indicating success/failure
        '''

        response = self._json_request(
            'POST',
            'abort_job/{0}/{1}'.format(namespace, job_id),
        )

        return response['aborted']

    def delete_job(self, namespace, job_id):
        '''
        Deletes a running Kubernetes job.

        Args:
            namespace (str): the namespace the job is running in
            job_id (str): the ID of the job to delete

        Returns:
            boolean indicating success/failure
        '''

        response = self._json_request(
            'POST',
            'delete_job/{0}/{1}'.format(namespace, job_id),
        )

        return response['deleted']



# Kubetools Client
# File: kubetools/cli/lists.py
# Desc:

import click
import six

from six.moves import xrange

from kubetools_client.cli import cli_bootstrap


@cli_bootstrap.group()
def list():
    '''
    List Kubetools things.
    '''


@list.command()
@click.pass_obj
def manifests(client):
    '''
    List manifests on the Kubetools server.
    '''

    click.echo('### Manifests')
    click.echo()

    for manifest in client.list_manifests():
        click.echo('--> {0}'.format(click.style(manifest, bold=True)))

    click.echo()


@list.command()
@click.pass_obj
def envs(client):
    '''
    List available Kubernetes environments.
    '''

    click.echo('### Environments')
    click.echo()

    for name, environment in six.iteritems(client.list_environments()):
        click.echo('--> {0}'.format(click.style(name, bold=True)))
        for key in ('host', 'proxy', 'kubernetes', 'registry'):
            click.echo('    {0}: {1}'.format(
                key.title(),
                click.style(environment[key], bold=True),
            ))

        click.echo()


@list.command()
@click.option('--status', default=None, help='Filter builds by status.')
@click.option('--limit', default=20, help='Number of builds to fetch.')
@click.option('--offset', default=0, help='Offset the builds to fetch.')
@click.pass_obj
def builds(client, status=None, limit=20, offset=0):
    '''
    List Kubetools builds.
    '''

    click.echo('### Builds')
    click.echo()

    builds = client.list_builds(
        status=status,
        limit=limit,
        offset=offset,
    )

    if not builds:
        click.echo('--> No builds found!')
        return

    # The API list builds most recent first, but in a CLI we actually want the most recent
    # at the bottom of the output.
    builds = sorted(builds, reverse=True)

    status_to_color = {
        'SUCCESS': 'green',
        'ERROR': 'red',
        'RUNNING': 'blue',
    }

    for build in builds:
        status = build['status']

        if status in status_to_color:
            status = click.style(status, status_to_color[status])

        click.echo('--> hash={0} started={1}, status={2}'.format(
            click.style(build['hash'], bold=True),
            build['date_started'], status,
        ))

    click.echo()


@list.command()
@click.argument('namespace', required=False)
@click.option('--print-pods', default=False, is_flag=True, help='Show pod names.')
@click.pass_obj
def jobs(client, namespace=None, print_pods=False):
    '''
    List jobs and namespaces in Kubernetes.
    '''

    click.echo('### Jobs')
    click.echo()

    namespace_jobs = client.list_jobs_by_namespace(namespace=namespace)

    for namespace, jobs in six.iteritems(namespace_jobs):
        click.echo('--> {0}'.format(click.style(namespace, bold=True)))
        click.echo('    Jobs:')

        for job_id, job in six.iteritems(jobs):
            status = '{0}/{1} completed'.format(
                job['succeeded'],
                job['completions'],
            )

            if job['succeeded'] == job['completions']:
                status = click.style(status, 'green')

            click.echo('    - {0}, {1}, {2} pods ({3})'.format(
                click.style(job['annotations'].get('description', 'Unknown'), bold=True),
                status,
                len(job['pods']),
                job_id,
            ))

            if print_pods:
                if job.get('pods'):
                    click.echo('      pods: {0}'.format(', '.join(
                        pod['name'] for pod in job['pods']
                    )))

        click.echo()


@list.command()
@click.argument('namespace', required=False)
@click.option(
    'print_urls', '--urls',
    default=False, is_flag=True,
    help='Show app URLs.',
)
@click.option(
    'print_deployments', '--deployments',
    default=False, is_flag=True,
    help='Show Kubernetes deployments.',
)
@click.option(
    'print_replicasets', '--replicasets',
    default=False, is_flag=True,
    help='Show Kubernetes replicasets.',
)
@click.option(
    'print_pods', '--pods',
    default=False, is_flag=True,
    help='Show Kubernetes pods.',
)
@click.option(
    '--kubernetes-objects',
    default=False, is_flag=True,
    help='Shortcut for --deployments --replicasets --pods',
)
@click.pass_obj
def apps(
    client, namespace,
    print_urls=False,
    print_deployments=False,
    print_replicasets=False,
    print_pods=False,
    kubernetes_objects=False,
):
    '''
    List apps and namespaces in Kubernetes.
    '''

    # Apply the --kubernetes-objects shortcut
    if kubernetes_objects:
        print_deployments = True
        print_replicasets = True
        print_pods = True

    click.echo('### Apps')
    click.echo()

    # Get the aps
    namespace_apps = client.list_apps_by_namespace(
        namespace=namespace,
        include_replicasets=print_replicasets,
        include_pods=print_pods,
    )

    # Figure out the longest service name for printing
    longest_name = 0
    name_lengths = []

    for apps in namespace_apps.values():
        if apps:
            name_lengths.extend(
                [len(name) for name in apps],
            )

    if name_lengths:
        longest_name = max(name_lengths) + 5

    # Now actually loop & click.echo(the output)
    for namespace, apps in six.iteritems(namespace_apps):
        click.echo('--> Namespace: {0}'.format(click.style(namespace, bold=True)))

        if apps:
            click.echo('    Apps:')

            for name, service in six.iteritems(apps):
                available_replicas = service['replicas']['available']
                desired_replicas = service['replicas']['desired']

                pod_status = '{0}/{1} running'.format(
                    available_replicas, desired_replicas,
                )
                version = service['annotations'].get('version')
                commit = service['selector'].get('commit')

                # Both version & commit (usual)
                if version and commit:
                    version = '{0} ({1})'.format(version, commit)

                # No version but commit
                elif not version and commit:
                    version = commit

                # No version, no commit
                elif not version:
                    version = 'unknown'

                if available_replicas == desired_replicas:
                    pod_status = click.style(pod_status, 'green')
                elif available_replicas > 0:
                    pod_status = click.style(pod_status, 'yellow')
                else:
                    pod_status = click.style(pod_status, 'red')

                click.echo('    - {0}{1}{2} @ {3}'.format(
                    click.style(name, bold=True),
                    ''.join(
                        ' ' for _ in xrange(
                            0, (longest_name - len(name)),
                        )
                    ),
                    pod_status, version,
                ))

                if print_urls and 'url' in service:
                    click.echo('      Link: {0}'.format(
                        click.style(service['url'][7:], bold=True),
                    ))

                kube_objects = service['kubernetes_objects']

                if print_deployments and 'deployment' in kube_objects:
                    click.echo('      deployment: {0}'.format(
                        click.style(kube_objects['deployment']['name']),
                    ))

                if print_replicasets and 'replicaset' in kube_objects:
                    click.echo('      Replicaset: {0}'.format(
                        click.style(kube_objects['replicaset']['name']),
                    ))

                if print_pods and 'pods' in kube_objects:
                    click.echo('      Pods:')
                    for pod in kube_objects['pods']:
                        click.echo('      - {0}'.format(pod['name']))

        click.echo()



from os import access, listdir, path, X_OK

import click

from pydash import memoize
from six.moves.configparser import ConfigParser

from kubetools_client.log import logger


class KubetoolsSettings(object):
    KUBETOOLS_HOST = None
    KUBETOOLS_PORT = 80
    KUBETOOLS_API_VERSION = 4

    DEFAULT_KUBE_ENV = 'staging'  # default environment when speaking to the server

    DEV_DEFAULT_ENV = 'dev'  # default environment name in dev
    DEV_HOST = 'localhost'  # dev host to link people to (should map to 127.0.0.1)
    DEV_CONFIG_DIRNAME = '.kubetools'  # project directroy to generate compose config

    KUBETOOLS_SESSION = None  # user auth/session details (generated by the server)

    def __init__(self, debug=False, filename=None):
        self.debug = debug
        self.filename = filename
        self.scripts = []


def get_settings_directory():
    return click.get_app_dir('kubetools', force_posix=True)


def _get_settings(debug=False):
    settings_directory = get_settings_directory()
    settings_file = path.join(settings_directory, 'kubetools.conf')

    settings = KubetoolsSettings(debug=debug, filename=settings_file)

    if path.exists(settings_file):
        logger.info('Loading settings file: {0}'.format(settings_file))
        parser = ConfigParser()
        parser.read(settings_file)

        for option in parser.options('kubetools'):
            setattr(
                settings,
                option.upper().replace('-', '_'),
                parser.get('kubetools', option),
            )

    else:
        logger.info('No settings file: {0}'.format(settings_file))

    scripts_directory = path.join(settings_directory, 'scripts')
    if path.exists(scripts_directory):
        for filename in listdir(scripts_directory):
            if access(path.join(scripts_directory, filename), X_OK):
                settings.scripts.append(filename)

    return settings


get_settings = memoize(
    _get_settings,
    # Ensure we always memoize even if not always called with the same args
    # this is because we initially bootstrap the settings at the CLI entry points
    # with the debug kwarg, but in sub-CLI functions we don't have the original
    # debug arg, but rely on the cached settings.
    resolver=lambda *args, **kwargs: 'settings',
)



#!/usr/bin/env python

from kubetools_client.dev import dev
from kubetools_client.main import run_cli

# Import click command groups
from kubetools_client.dev import container, environment, logs, scripts  # noqa: F401, I100


run_cli(dev)



import click

from kubetools_client import __version__
from kubetools_client import KubeClient
from kubetools_client.log import setup_logging
from kubetools_client.settings import get_settings


@click.group()
@click.option(
    '-s', '--server',
    envvar='KUBETOOLS_HOST',
)
@click.option(
    '-p', '--port',
    type=int,
    envvar='KUBETOOLS_PORT',
)
@click.option('kube_env', '--env', '--kube-env')
@click.option('--debug', is_flag=True)
@click.version_option(version=__version__, message='%(prog)s: v%(version)s')
@click.pass_context
def cli_bootstrap(ctx, server=None, port=None, kube_env=None, debug=False):
    '''
    Kubetools client.
    '''

    setup_logging(debug)

    # Get settings
    settings = get_settings(debug)

    # CLI options > settings
    host = server or settings.KUBETOOLS_HOST
    port = port or settings.KUBETOOLS_PORT
    kube_env = kube_env or settings.DEFAULT_KUBE_ENV

    # Setup the client
    ctx.obj = KubeClient(host=host, port=port, kube_env=kube_env)



# Kubetools Client
# File: kubetools/cli/__init__.py
# Desc: helpers for the Kubetools command line interface

import click
import six

from kubetools_client.cli import cli_bootstrap
from kubetools_client.cli.wait import wait_build, wait_job
from kubetools_client.exceptions import KubeCLIError
from kubetools_client.log import logger


def _get_apps_in_namespace(client, namespace):
    return client.list_apps_by_namespace().get(namespace, {})


def _parse_envars(envars):
    return dict(
        envar.split('=', 1)
        for envar in envars
    )


def _parse_build_options(apps, envars=None):
    apps_to_build = []

    for app in apps:
        replicas = version = None

        if '*' in app:
            app, replicas = app.split('*')
            replicas = int(replicas)

        if '=' in app:
            app, version = app.split('=')

        app_to_build = {
            'app': app,
        }

        if replicas:
            app_to_build['replicas'] = replicas

        if version:
            app_to_build['version'] = version

        if envars:
            app_to_build['envars'] = _parse_envars(envars)

        apps_to_build.append(app_to_build)

    logger.debug('Converted CLI app spec: {0} -> {1}'.format(
        apps, apps_to_build,
    ))

    return apps_to_build


def _generate_build_deploy(
    action, client, func,
    build_options=None,
    envars=None,
    **kwargs
):
    '''
    CLI helper for generating builds/deploys from command line input and then waiting on
    Kubetools to complete the build.
    '''

    build_options = build_options or []
    namespace = kwargs.get('namespace')

    click.echo('### {0}: {1}{2}'.format(
        action,
        '{0}/'.format(namespace) if namespace else '',
        click.style((
            ', '.join(build_options)
            if build_options
            else 'all'
        ), bold=True),
    ))
    click.echo()

    apps = _parse_build_options(build_options, envars=envars)

    # Cleanup only provides namespace
    if func == client.create_cleanup:
        build_hash = func(kwargs.pop('namespace'))

    # Upgrade/remove require namespace first
    elif func in (client.create_upgrade, client.create_remove, client.create_restart):
        build_hash = func(kwargs.pop('namespace'), apps)

    # Deploy/build doesn't need a namespace
    else:
        build_hash = func(apps, **kwargs)

    # Now wait for the build to complete
    click.echo('--> {0} queued, hash: {1}'.format(action, click.style(build_hash, bold=True)))
    wait_build(client, build_hash)


def _check_existing_apps(client, namespace, action):
    # Build full app spec based on the existing services
    existing_apps = _get_apps_in_namespace(client, namespace)

    if not existing_apps:
        raise KubeCLIError('No apps under namespace {0}, nothing to {1}'.format(
            namespace, action,
        ))


@cli_bootstrap.command()
@click.argument('build_options', nargs=-1, required=True)
@click.pass_obj
def build(client, build_options):
    '''
    Builds all containers needed for an app.
    '''

    return _generate_build_deploy(
        'Building', client, client.create_build, build_options,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.argument('deploy_options', nargs=-1, required=True)
@click.option(
    '-e', '--envar',
    multiple=True,
    help='Envars to pass into the app containers.',
)
@click.pass_obj
def deploy(client, deploy_options, namespace=None, envar=None):
    '''
    Deploy apps into a new or existing namespace.
    '''

    return _generate_build_deploy(
        'Deploying', client, client.create_deploy,
        build_options=deploy_options,
        namespace=namespace,
        envars=envar,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.argument('upgrade_options', nargs=-1)
@click.pass_obj
def upgrade(client, namespace, upgrade_options):
    '''
    Upgrade apps in a namespace.
    '''

    _check_existing_apps(client, namespace, 'upgrade')

    return _generate_build_deploy(
        'Upgrading', client, client.create_upgrade,
        build_options=upgrade_options,
        namespace=namespace,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.argument('apps_to_remove', nargs=-1)
@click.pass_obj
def remove(client, namespace, apps_to_remove):
    '''
    Remove apps from a namespace.
    '''

    _check_existing_apps(client, namespace, 'remove')

    return _generate_build_deploy(
        'Removing', client, client.create_remove,
        build_options=apps_to_remove,
        namespace=namespace,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.argument('apps_to_restart', nargs=-1)
@click.pass_obj
def restart(client, namespace, apps_to_restart):
    '''
    Restart apps in a namespace.
    '''

    _check_existing_apps(client, namespace, 'restart')

    return _generate_build_deploy(
        'Restarting', client, client.create_restart,
        build_options=apps_to_restart,
        namespace=namespace,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.pass_obj
def cleanup(client, namespace):
    '''
    Cleanup a namespace.
    '''

    _check_existing_apps(client, namespace, 'cleanup')

    return _generate_build_deploy(
        'Cleaning', client, client.create_cleanup,
        namespace=namespace,
    )


@cli_bootstrap.command()
@click.argument('namespace')
@click.argument('app')
@click.argument('command')
@click.option(
    '-e', '--envar',
    multiple=True,
    help='Envars to pass into the app containers.',
)
@click.option(
    '--chdir',
    help='Directory inside the container to execute the command.',
)
@click.pass_obj
def run(
    client, namespace, app, command,
    envar=None, chdir=None,
):
    '''
    Run commands for an app in a given namespace.
    '''

    click.echo('### Running {0} on {1}'.format(
        click.style(command, bold=True),
        click.style(app, bold=True),
    ))
    click.echo()

    apps = _parse_build_options([app])
    job_spec = {
        'command': command,
    }

    if chdir:
        job_spec['chdir'] = chdir

    # Parse/inject any envars
    if envar:
        job_spec['envars'] = _parse_envars(envar)

    # Splice in our command for each app as a job
    for app in apps:
        app['job'] = job_spec

    build_hash = client.create_run(apps, namespace=namespace)

    # Now wait for the *build* to complete
    click.echo('--> Waiting for build: {0}'.format(click.style(build_hash, bold=True)))
    wait_build(client, build_hash)

    # Now fetch the jobs in the namespace and find the job created by this build
    namespace_jobs = client.list_jobs_by_namespace(namespace=namespace)
    jobs = namespace_jobs[namespace]

    job_id = None

    for id_, data in six.iteritems(jobs):
        if data['annotations'].get('kubetools/build_hash') == build_hash:
            job_id = id_
            break

    click.echo()
    click.echo('--> Waiting for job: {0}'.format(click.style(job_id, bold=True)))
    wait_job(client, namespace, job_id)
